use warnings;
use strict;

package Mozilla::Persona::Server;
use open 'utf8';

use Log::Report    qw/persona/;

use Crypt::OpenSSL::Bignum ();
use Crypt::OpenSSL::RSA    ();

use CGI::Fast      ();
use CGI::Session   ();
use JSON           qw(decode_json encode_json);
use File::Slurp    qw/read_file write_file/;
use MIME::Base64   qw(encode_base64url);
use Time::HiRes    qw(time);
use List::Util     qw(first);
use File::Basename qw(dirname);
use File::Spec     ();

use constant MIME_JSON => 'application/json; charset=UTF-8';

=chapter NAME
Mozilla::Persona::Server - persona user validation

=chapter SYNOPSIS
  # run set-up script first
  persona setup <options>

  my $persona = Mozilla::Persona->new(...);
  my $persona = Mozilla::Persona->fromConfig($fn);
  $persona->writeConfig($fn);

=chapter DESCRIPTION
This module organizes the server-side Persona adminstration, with
a few helpers which can have alternative implementations.

=chapter METHODS

=section Constructors

=c_method new OPTIONS
=requires private_pem FILENAME
FILENAME which contains the private key, generated by the set-up
script.

=requires cookie_name STRING
The name of the cookie, used to administer the persona session.

=requires domain DOMAIN
Your domain, which may be a super-level from the website's domain,
or even an other domain you own.

=option  aliases HASH
=default aliases See manpage M<Mozilla::Persona::Aliases>
One used may have more than one valid email address for various
purposes.  Any of these aliases suffices to login.  The HASH of
information is passed as configuration for the ::Aliases object.

=option  validator HASH
=default validator See manpage M<Mozilla::Persona::Validate>
Used to check whether the user is validated for this domain.
The HASH of information is passed as configuration for the
::Aliases object.
=cut

sub new(%) { my $class = shift; (bless {}, $class)->init({@_}) }
sub init($)
{   my ($self, $args) = @_;
    $self->{MP_pem_fn}   = $args->{private_pem} or panic;
    $self->{MP_cookie}   = $args->{cookie_name} or panic;
    $self->{MP_domain}   = $args->{domain}      or panic;
    $self->{MP_aliparms} = $args->{aliases}
       || {class => 'Mozilla::Persona::Aliases'};
    $self->{MP_valparms} = $args->{validator}
       || {class => 'Mozilla::Persona::Validate::Table'};
    $self;
}

=c_method fromConfig FILENAME, OPTIONS
Read the configuration from FILENAME (which contains JASON).  The OPTIONS
overrule the read data.  The data is used to instantiate this object via
the usual procedure, so see M<new()>.
=cut

sub fromConfig($)
{   my ($class, $fn) = (shift, shift);
    my $config = decode_json read_file $fn;
    $class->new(%$config, @_);
}

#-----------------
=section Accessors
=method cookie
=method domain
=cut

sub cookie()  {shift->{MP_cookie}}
sub domain()  {shift->{MP_domain}}

=method aliases
Returns the aliases object, an M<Mozilla::Persona::Aliases> extension.

The object gets instantiated on first call to this method the first call
to this method, so lazy.
=cut

sub aliases()
{   my $self = shift;
    return $self->{MP_aliases}
        if $self->{MP_aliases};

    my $config = $self->{MP_aliparms} || {};

    # load alias expansion plugin
    my $class = delete $config->{class} or panic;
    eval {require $class}; panic $@ if $@;

    $self->{MP_aliases} = $class->new(%$config);
}

=method validator
Returns the validator object, an M<Mozilla::Persona::Validate> extension.

The object gets instantiated on first call to this method the first call
to this method, so lazy.
=cut

sub validator()
{   my $self = shift;
    return $self->{MP_validator}
        if $self->{MP_validator};

    my $config = $self->{MP_valparms} || {};

    # load username/password validator
    my $class = delete $config->{class} or panic;
    eval {require $class}; panic $@ if $@;

    $self->{MP_validator} = $class->new(%$config);
}

=method privatePEM
Returns the M<Crypt::OpenSSL::RSA> object which contains the
private PEM (private key).
=cut

sub privatePEM()
{   my $self = shift;

    my $pem  = read_file $self->{MP_pem_fn};
    my $key  = Crypt::OpenSSL::RSA->new_private_key($pem);
    $key->use_pkcs1_padding;
    $key->use_sha256_hash;
    $key;
}

#------------------------
=section Actions
Actions are activities requested by the client.

=method getSession CGI
Return the M<CGI::Session> for this persona user, based of information
received via M<CGI>.
=cut

sub getSession($)
{   my ($self, $cgi) = @_;
    my $cookie  = $cgi->cookie($self->cookie)
        or error __x"no session cookie";

    my $session = CGI::Session->new('driver:File', $cookie)
        or error __x"invalid session cookie";

    $session;
}

=method actionSign CGI
Handle a sign request.
=cut

sub _sign($$$)
{   my ($self, $client_pubkey, $email, $duration) = @_;

   # NB.  Treating the jwcrypto code as the spec here.
    my $issued_at = int(1000*time);

    my %cert      =
      ( iss          => $self->domain
      , exp          => $issued_at + 1000*$duration
      , iat          => $issued_at
      , "public-key" => $client_pubkey
      , principal    => { email => $email }
      );

    my %header     =
      ( typ          => 'JWT'
      , alg          => 'RS256'
      );

    my $header_enc = encode_base64url encode_json \%header;
    my $cert_enc   = encode_base64url encode_json \%cert;

    my $key       = $self->privatePEM or return;
    my $sig_enc    = encode_base64url $key->sign("$header_enc.$cert_enc");

    "$header_enc.$cert_enc.$sig_enc";
}

sub actionSign($)
{   my ($self, $cgi) = @_;

    my $session  = $self->getSession($cgi);
    my $user     = $session->param('user');

    print $cgi->header(-content_type => MIME_JSON);

    my $user_pubkey  = $cgi->param('pubkey')
        or error __x"nothing to sign for {user}", user => $user;

    my $duration = $cgi->param('duration') || 24*3600;

    my $email        = $cgi->param('email')
        or error __x"no email address to sign for {user}", user => $user;

    $self->isAliasFor($user, $email)
        or error __x"user {username} is not authorized to use {email}"
            , username => $user, email => $email;
         
    trace "signed $user $email";

    my $sig = $self->_sign(decode_json($user_pubkey), $email, $duration);
    print encode_json({signature => $sig}), "\n";
}

=method actionLogin CGI
Handle a login request.
=cut

sub actionLogin($)
{   my ($self, $cgi) = @_;

    my $session;
    if(my $cookie  = $cgi->cookie($self->cookie))
    {   $session = CGI::Session->new("driver:File", $cookie);
    }

    if($session)
    {   # session restored
        print $cgi->header(-content_type => MIME_JSON);
    }
    else
    {   # new session, new cookie
        $session = CGI::Session->("driver:File", undef);
        my $cookie = $cgi->cookie
          ( -name     => $self->cookie
          , -value    => $session->id
          , -expires  => '+1d'
          , -secure   => 1
          , -httponly => 1
          , -domain   => $self->domain
          );
        print $cgi->header(-content_type => MIME_JSON, -cookie => $cookie);
    }

    my $email    = $cgi->param('email')
        or error __x"no email address provided";

    my $password = $cgi->param('password')
        or error __x"no password provided for {email}", email => $email;

    my $validator = $self->validator;
    my @aliases   = $self->aliases->for($email);
    my $user      = first {$validator->isValid($_, $password)} @aliases;

    if($user)
    {   trace "authenticated $user";
        $session->param(user => $user);
        print encode_json({user => $user}), "\n";
    }
    else
    {   error __x"authentication for {email} failed (aliases {aliases})"
          , email => $email, aliases => \@aliases;
    }
}

=method actionIsLoggedIn CGI
Check whether the user is logged-in.
=cut

sub actionIsLoggedIn($)
{   my ($self, $cgi) = @_;

    my $is_logged_in = 0;
    if(my $cookie  = $cgi->cookie($self->cookie))
    {   my $email  = $cgi->param('email');
        if(my $session = CGI::Session->new('driver:File', $cookie))
        {   my $user   = $session->param('user');
            $is_logged_in = $email && $user && $self->isAliasFor($user, $email);
            trace "$email $user is ".($is_logged_in ? '' : ' not')."logged in";
        }
        else
        {   trace "no session for $email";
        }
    }
    else
    {   trace "not logged in";
    }

    print $cgi->header(-content_type => MIME_JSON)
        , encode_json({logged_in_p => $is_logged_in})
        , "\n";
}

#------------------------------
=section Helpers

=method isAliasFor USER, EMAIL
Check whether the USER is one of the aliases for the specified EMAIL
address.
=cut

sub isAliasFor($$)
{   my ($self, $user, $email) = @_;

    my @aliases  = $self->aliases->for($email);
    @aliases
        or error __x"user {user} for {email} not found"
            , user => $user, email => $email;

    my $lc_user = lc $user;
    first {$lc_user eq lc $_} @aliases;
}

=method writeConfig FILENAME
Write the configuration of this object to the FILENAME.
=cut

sub writeConfig($)
{   my ($self, $fn) = @_;
    my %data;
 
    @data{ qw/private_pem cookie_name domain aliases validator/ }
      = @{$self}{ qw/MP_pem_fn MP_cookie MP_domain MP_aliparms MP_valparms/ };

    write_file $fn, JSON->new->utf8->pretty(1)->encode(\%data);
}

1;
