
use warnings;
use strict;

package Mozilla::Persona::Validate::Htpasswd;
use base 'Mozilla::Persona::Validate';

use Log::Report    qw/persona/;
use Apache::Htpasswd ();

=chapter NAME
Mozilla::Persona::Validate::Htpasswd - validate the existence via an Apache Htpasswd file

=chapter SYNOPSIS
  # add to the configation of $domain.json
  { ...
    "validate" : {
      "class"  : "Mozilla::Persona::Validator::Htpasswd",
      "pwfile" : "$path_to_file",
    },
    ...
  }

=chapter DESCRIPTION
The validator object must ensure that the user/password combination
is valid.  In this case, an htpasswd file (as defined by Apache) is
used to authenticate the user.  This file can be shared between the
website configuration and Persona, but does not need to.

You can use apache's C<htpasswd> or C<htpasswd2> command to maintain
the file.

=chapter METHODS

=section Constructors

=c_method new OPTIONS
=requires pwfile FILENAME
The htpasswd file.  Probably, you want to provide an absolute path.
When the password file changes, it will automatically be reloaded.
=cut

sub init($)
{   my ($self, $args) = @_;

    my $fn = $args->{pwfile} or panic;
    $self->openFile($fn);   # pre-load
    $self;
}

#------------
=section Accessors
=method pwfile
=cut

sub pwfile() {shift->{MPVH_fn}}

sub openFile(;$)
{   my ($self, $fn) = @_;

    if($fn) { $self->{MPVH_fn} = $fn }
    else    { $fn = $self->{MPVH_fn} }

    my $mtime = (stat $fn)[9];
    defined $mtime
        or fault __x"htpasswd file {fn}";

    if(my $last_mtime = $self->{MPVH_mtime})
    {   return $self->{MPVH_info} if $mtime eq $last_mtime;
    }

    my $info = $self->{MPVH_info} = Apache::Htpasswd
      ->new({passwdFile => $fn, ReadOnly => 1});

    $self->{MPVH_mtime} = $mtime;
    $info;
}

sub isValid($$)
{   my ($self, $user, $password) = @_;
    $self->openFile->htCheckPassword($user, $password);
}

1;
